#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <GL/glew.h>
#include <SDL2/SDL.h>
#include "app.h"

static bool init(int argc, char **argv);
static void process_event(SDL_Event *ev);
static void proc_modkeys();

static SDL_Window *win;
static SDL_GLContext ctx;
static bool fullscreen, mouse_grabbed;
static bool quit;

static unsigned int start_time;
static unsigned int modkeys;

SDL_GameController *gamepad;

static int scale_factor = 1;

int main(int argc, char **argv)
{
	if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_GAMECONTROLLER) != 0) {
		fprintf(stderr, "failed to initialize SDL\n");
		return 1;
	}

	SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
	SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, 8);
	SDL_GL_SetAttribute(SDL_GL_FRAMEBUFFER_SRGB_CAPABLE, 1);

	int defpos = SDL_WINDOWPOS_UNDEFINED;
	unsigned int sdlflags = SDL_WINDOW_OPENGL | SDL_WINDOW_RESIZABLE | SDL_WINDOW_ALLOW_HIGHDPI;

	if(!(win = SDL_CreateWindow("demo", defpos, defpos, 1024, 768, sdlflags))) {
		// try again without sRGB capability
		SDL_GL_SetAttribute(SDL_GL_FRAMEBUFFER_SRGB_CAPABLE, 0);
		if(!(win = SDL_CreateWindow("demo", defpos, defpos, 1024, 768, sdlflags))) {
			fprintf(stderr, "failed to create window\n");
			SDL_Quit();
			return 1;
		}
		fprintf(stderr, "failed to get an sRGB framebuffer.\n");
	}
	int val;
	SDL_GL_GetAttribute(SDL_GL_FRAMEBUFFER_SRGB_CAPABLE, &val);
	printf("SDL says we %s an sRGB framebuffer\n", val ? "got" : "didn't get");

	if(!(ctx = SDL_GL_CreateContext(win))) {
		fprintf(stderr, "failed to create OpenGL context\n");
		SDL_Quit();
		return 1;
	}
	SDL_GL_GetDrawableSize(win, &win_width, &win_height);
	win_aspect = (float)win_width / (float)win_height;

	printf("detected %d joysticks\n", SDL_NumJoysticks());
	for(int i=0; i<SDL_NumJoysticks(); i++) {
		if(SDL_IsGameController(i)) {
			if(!(gamepad = SDL_GameControllerOpen(i))) {
				fprintf(stderr, "failed to open game controller %i: %s\n", i, SDL_GetError());
				continue;
			}
			printf("Using gamepad: %s\n", SDL_GameControllerNameForIndex(i));
		}
	}

	if(!init(argc, argv)) {
		SDL_Quit();
		return 1;
	}
	app_reshape(win_width, win_height);

	while(!quit) {
		SDL_Event ev;

		time_msec = SDL_GetTicks() - start_time;
		while(SDL_PollEvent(&ev)) {
			process_event(&ev);
			if(quit) goto break_evloop;
		}

		app_display();
	}
break_evloop:

	app_cleanup();
	SDL_Quit();
	return 0;
}

void app_swap_buffers()
{
	SDL_GL_SwapWindow(win);
}

void app_quit()
{
	quit = true;
}

unsigned int app_get_modifiers()
{
	return modkeys;
}

void app_resize(int x, int y)
{
	SDL_SetWindowSize(win, x, y);
}

void app_fullscreen(bool fs)
{
	SDL_SetWindowFullscreen(win, fs ? SDL_WINDOW_FULLSCREEN_DESKTOP : 0);
	fullscreen = fs;
}

void app_toggle_fullscreen()
{
	app_fullscreen(!fullscreen);
}

bool app_is_fullscreen()
{
	return fullscreen;
}

void app_grab_mouse(bool grab)
{
	if(grab) {
		SDL_WarpMouseInWindow(win, win_width / 2, win_height / 2);
	}
	//SDL_SetWindowGrab(win, grab ? SDL_TRUE : SDL_FALSE);
	//SDL_ShowCursor(grab ? 1 : 0);
	SDL_SetRelativeMouseMode(grab ? SDL_TRUE : SDL_FALSE);
	mouse_grabbed = grab;
}

void app_toggle_grab_mouse()
{
	app_grab_mouse(!mouse_grabbed);
}

bool app_is_mouse_grabbed()
{
	return mouse_grabbed;
}


static bool init(int argc, char **argv)
{
	glewInit();

	if(!app_init(argc, argv)) {
		return false;
	}

	start_time = SDL_GetTicks();
	return true;
}

static void process_event(SDL_Event *ev)
{
	switch(ev->type) {
	case SDL_QUIT:
		quit = true;
		break;

	case SDL_KEYDOWN:
	case SDL_KEYUP:
		proc_modkeys();
		app_keyboard(ev->key.keysym.sym, ev->key.state == SDL_PRESSED);
		break;

	case SDL_MOUSEBUTTONDOWN:
	case SDL_MOUSEBUTTONUP:
		proc_modkeys();
		app_mouse_button(ev->button.button - SDL_BUTTON_LEFT, ev->button.state == SDL_PRESSED,
				ev->button.x * scale_factor, ev->button.y * scale_factor);
		break;

	case SDL_MOUSEMOTION:
		if(mouse_grabbed) {
			app_mouse_delta(ev->motion.xrel, ev->motion.yrel);
		} else {
			app_mouse_motion(ev->motion.x * scale_factor, ev->motion.y * scale_factor);
		}
		break;

	case SDL_WINDOWEVENT:
		if(ev->window.event == SDL_WINDOWEVENT_RESIZED) {
			SDL_GL_GetDrawableSize(win, &win_width, &win_height);
			win_aspect = (float)win_width / (float)win_height;
			scale_factor = win_width / ev->window.data1;
			app_reshape(win_width, win_height);
		}
		break;

	case SDL_CONTROLLERAXISMOTION:
		app_gamepad_axis(ev->caxis.axis, ev->caxis.value / 32768.0f);
		break;

	case SDL_CONTROLLERBUTTONDOWN:
	case SDL_CONTROLLERBUTTONUP:
		app_gamepad_button(ev->cbutton.button, ev->type == SDL_CONTROLLERBUTTONDOWN);
		break;
	}
}

static void proc_modkeys()
{
	modkeys = 0;
	SDL_Keymod sdlmod = SDL_GetModState();
	if(sdlmod & KMOD_SHIFT) {
		modkeys |= MOD_SHIFT;
	}
	if(sdlmod & KMOD_ALT) {
		modkeys |= MOD_ALT;
	}
	if(sdlmod & KMOD_CTRL) {
		modkeys |= MOD_CTRL;
	}
}
